# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Shared code for all scripts and modules """

import logging

log = logging.getLogger("dlab.core")


def get_or_verify_datafile(resource, skip_verify=False):
    """Load a neurobank resource, either from a local file or from an archive

    if `resource` is a file that exists, the hash is used to look up resource information
    if `resource` is an ID or a registry URL, tries to find the file in a local archive

    If `skip_verify` is True, does the lookup by the name of the file (to avoid long hashes)

    returns (datafile path, registry record)
    """
    from pathlib import Path
    import nbank

    p = Path(resource)
    if p.exists():
        datafile = p
        if skip_verify:
            log.info(" - using local file %s", datafile)
            name = p.stem
            resource_info = nbank.describe(name)
        else:
            log.info(" - using local file %s (checking hash)", datafile)
            resources = nbank.verify(datafile)
            try:
                resource_info = next(resources)
                log.info("   ✓ %s", resource_info["sha1"])
            except StopIteration:
                raise ValueError(
                    "sha1 for local file does not match any resource in the registry"
                )
    else:
        resource_url = nbank.full_url(resource)
        log.info(" - source resource: %s", resource_url)
        resource_info = nbank.describe(resource_url)
        datafile = nbank.get(resource, local_only=True)
        if datafile is None:
            raise ValueError("unable to locate resource file")
    return datafile, resource_info


def fetch_resource(neurobank_registry, resource_name):
    """Fetch a downloadable resource from the registry.

    The file will be cached locally. Returns the path of the file.
    """
    from pathlib import Path
    from appdirs import user_cache_dir
    from urllib.parse import urlparse
    from nbank import fetch_resource

    APP_NAME = "dlab"
    APP_AUTHOR = "melizalab"
    parsed_url = urlparse(neurobank_registry)
    cache_dir = Path(user_cache_dir(APP_NAME, APP_AUTHOR)) / parsed_url.netloc
    cache_dir.mkdir(parents=True, exist_ok=True)
    target = cache_dir / resource_name
    if not target.exists():
        log.debug("   - fetching %s from registry", resource_name)
        fetch_resource(neurobank_registry, resource_name, target)
    return str(target)



# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Functions for using mountainlab data """
import os
import shutil
import logging

from dlab import core, __version__

log = logging.getLogger('dlab.mountain')


def assign_events(pprox, events):
    """Assign events to trials within a pprox based on recording time.

    trials: an iterable list of pproc objects, sorted in order of time. Each
    object must have a "recording" field that contains "start" and "stop"
    subfields. The values of these fields must indicate the start and stop time
    of the trial.

    """
    from dlab.pprox import trial_iterator
    from copy import deepcopy

    clusters = {}
    trial_iter = trial_iterator(pprox)
    index, trial = next(trial_iter)
    for channel, time, clust in events:
        if time < trial["recording"]["start"]:
            log.warning("warning: spike at %d is before the start of trial %d", time, index)
            continue
        while time > trial["recording"]["stop"]:
            try:
                index, trial = next(trial_iter)
            except StopIteration:
                log.warning(
                    "warning: spike at %d is after the end of the last trial (%d samples)",
                    time,
                    trial["recording"]["stop"],
                )
                break
        t_seconds = (time - trial["recording"]["start"]) / trial["recording"]["sampling_rate"]
        if clust not in clusters:
            cluster = deepcopy(pprox)
            cluster.update(cluster=clust, channel=channel)
            clusters[clust] = cluster
        clusters[clust]["pprox"][index]["events"].append(t_seconds)
    return clusters


def group_spikes_script(argv=None):
    import nbank
    import argparse
    import json
    from arfx import mdaio
    from dlab.util import setup_log, json_serializable
    __version__ = "0.1.0"

    p = argparse.ArgumentParser(
        description="group sorted spikes into pprox files based on cluster and trial"
    )
    p.add_argument(
        "-v", "--version", action="version", version="%(prog)s " + __version__
    )
    p.add_argument("--debug", help="show verbose log messages", action="store_true")
    p.add_argument(
        "--output",
        "-o",
        help="directory to output pprox files (default current directory)",
    )
    p.add_argument(
        "--name",
        "-n",
        help="base name of the unit (default is based on 'recording' field of trials pprox) ",
    )
    p.add_argument("trials", help="pprox file with the trial structure of the experiment")
    p.add_argument("firings", help="firings.mda file generated by mountainsort")
    args = p.parse_args(argv)
    setup_log(log, args.debug)

    log.info("- loading data:")
    log.info("  - experiment file: %s", args.trials)
    with open(args.trials, "rt") as fp:
        pprox = json.load(fp)
    log.info("  - spike times: %s", args.firings)
    with mdaio.mdafile(args.firings) as fp:
        events = fp.read().astype("i8")

    if args.name is None:
        base, rec_id = nbank.parse_resource_id(pprox["recording"])
        args.name = rec_id

    log.info("- grouping spikes by cluster and trial...")
    clusters = assign_events(pprox, events)
    for clust_id, cluster in clusters.items():
        outfile = os.path.join(args.output or "", "{}_c{}.pprox".format(args.name, clust_id))
        log.info("  - cluster %d -> %s", clust_id, outfile)
        try:
            pb = cluster["processed_by"]
        except KeyError:
            pb = []
            cluster["processed_by"] = pb
        pb.append("{} {}".format(p.prog, __version__))
        with open(outfile, "wt") as ofp:
            json.dump(cluster, ofp, default=json_serializable)



# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Functions for using kilsort/phy data """
import logging
import json
from pathlib import Path
from collections import namedtuple

import numpy as np
import pandas as pd
import quickspikes as qs
import h5py as h5
from dlab import pprox

log = logging.getLogger("dlab.kilo")

Trial = namedtuple(
    "Trial",
    [
        "recording_entry",
        "recording_start",
        "recording_end",
        "stimulus_name",
        "stimulus_start",
        "stimulus_end",
    ],
)

Stimulus = namedtuple("Stimulus", ["name", "start", "end"], defaults=[None])


def read_kilo_params(fname):
    """Read the kilosort params.py file"""
    from itertools import chain
    from configparser import ConfigParser

    parser = ConfigParser()
    with open(fname, "rt") as lines:
        lines = chain(("[top]",), lines)
        parser.read_file(lines)
    sect = parser["top"]
    return dict(
        dtype=sect["dtype"].strip("'"),
        nchannels=int(sect["n_channels_dat"]),
        sampling_rate=float(sect["sample_rate"]),
    )


def oeaudio_stims(dset):
    """Parse the messages in the 'stim' dataset to get a table of stimuli with
    start samples. Note that these need to be corrected for offset of the
    recording and network lag.

    """
    import re

    re_start = re.compile(r"start (.*)")

    for row in dset:
        time = row["start"]
        message = row["message"].decode("utf-8")
        m = re_start.match(message)
        if m is not None:
            stim_name = Path(m.group(1)).stem
            yield Stimulus(stim_name, time)


def oeaudio_to_trials(data_file, sync_dset, sync_thresh=1.0, prepad=1.0):

    """Extracts trial information from an oeaudio-present experiment ARF file

    When using oeaudio-present, a single recording is made in response to all
    the stimuli. The stimulus presentation script sends network events to
    open-ephys to mark the start and stop of each stimulus. There is typically a
    significant lag between the 'start' event and the onset of the stimulus, due
    to buffering of the audio playback. However, the oeaudio-present script will
    play a synchronization click on a second channel by default. As long as the
    user remembers to record this channel, it can be used to correct the onset
    and offset values.

    The continuous recording is broken up into trials based on the stimulus
    presentation, such that each trial encompasses one and only one stimulus.
    The `prepad` parameter specifies, in seconds, when trials begin relative to
    stimulus onset. The default is 1.0 s.

    """
    from itertools import zip_longest
    from dlab.extracellular import (
        find_stim_dset,
        entry_time,
        iter_entries,
        stim_duration,
    )
    from arf import timestamp_to_datetime

    expt_start = None
    det = qs.detector(sync_thresh, 10)
    for entry_num, entry in iter_entries(data_file):
        log.info(" - entry: '%s'", entry.name)
        entry_start = entry_time(entry)
        log.info("  - start time: %s", timestamp_to_datetime(entry.attrs["timestamp"]))
        if expt_start is None:
            expt_start = entry_start

        log.info("  - parsing stimulus log")
        stims = list(oeaudio_stims(find_stim_dset(entry)))
        log.info("    - detected %d stimuli", len(stims))
        log.info("  - sync track: '%s'", sync_dset)
        sync = entry[sync_dset]
        sync_data = sync[:].astype("d")
        det.scale_thresh(sync_data.mean(), sync_data.std())
        stim_onsets = np.asarray(det(sync_data))
        log.info("    - detected %d clicks", stim_onsets.size)
        dset_offset = sync.attrs["offset"]
        dset_end = sync.size
        sampling_rate = sync.attrs["sampling_rate"]
        stim_sample_offset = int(dset_offset * sampling_rate)
        log.info("  - recording clock offset: %d", stim_sample_offset)

        if len(stims) != stim_onsets.size:
            log.warning(
                "  - WARNING: number of stimuli does not match number of clicks. This recording may need to be discarded"
            )

        padding_samples = int(prepad * sampling_rate)
        for stim, onset, offset in zip_longest(
            stims, stim_onsets, stim_onsets[1:], fillvalue=dset_end + padding_samples
        ):
            stim_dur = int(stim_duration(stim.name) * sampling_rate)
            if stim_dur > offset - onset:
                log.warning(
                    "  - WARNING: stimulus %s is longer than the duration of the trial",
                    stim,
                )
            yield Trial(
                entry_num,
                onset - padding_samples,
                offset - padding_samples,
                stim.name,
                onset,
                onset + stim_dur,
            )


def assign_events_flat(events, sampling_rate):
    """Assign event_times to clusters, generating a large toelis object"""
    nevents, _ = events.shape
    log.info("- grouping %d spikes by cluster...", nevents)
    return (
        events
        .groupby("clust")
        .apply(lambda df: df.time.sort_values().to_numpy() / sampling_rate * 1000.0)
    )


def trials_to_pprox(trials, sampling_rate):
    """ Convert pandas trials to pproc """
    for trial in trials.itertuples():
        pproc = {
            "events": (trial.events.astype("d") - trial.stimulus_start) / sampling_rate,
            "offset": trial.stimulus_start / sampling_rate,
            "index": trial.Index,
            "interval": ((trial.recording_start - trial.stimulus_start) / sampling_rate,
                         (trial.recording_end - trial.stimulus_start) / sampling_rate),
            "stimulus": {
                "name": trial.stimulus_name,
                "interval": (0.0,
                             (trial.stimulus_end - trial.stimulus_start) / sampling_rate)
            },
            "recording": {
                "entry": trial.recording_entry,
                "start": trial.recording_start,
                "end": trial.recording_end
            }
        }
        yield pproc


def group_spikes_script(argv=None):
    import nbank
    import argparse
    from dlab.util import setup_log, json_serializable
    from dlab.core import get_or_verify_datafile
    from dlab.extracellular import entry_metadata, iter_entries

    __version__ = "2022.10.11"

    p = argparse.ArgumentParser(
        description="group kilosorted spikes into pprox files based on cluster and trial"
    )
    p.add_argument(
        "-v", "--version", action="version", version="%(prog)s " + __version__
    )
    p.add_argument("--debug", help="show verbose log messages", action="store_true")
    p.add_argument(
        "--dry-run",
        help="do everything except write the output files",
        action="store_true",
    )
    p.add_argument(
        "--sync",
        default="sync",
        help="name of channel with synchronization signal (default '%(default)s')",
    )
    p.add_argument(
        "--sync-thresh",
        default=30.0,
        type=float,
        help="threshold (z-score) for detecting sync clicks (default %(default)0.1f)",
    )
    p.add_argument(
        "--prepad",
        type=float,
        default=1.0,
        help="sets trial start time relative to stimulus onset (default %(default)0.1f s)",
    )
    p.add_argument(
        "--toelis",
        action="store_true",
        help="output toelis instead of pprox. one file will be generated for "
        "the entire recording (including multiunits)",
    )
    p.add_argument(
        "--cluster",
        "-c",
        help="only save data for the specified clusters (as comma-separated list)",
        type=lambda s: [int(item) for item in s.split(",")],
    )
    p.add_argument(
        "--output",
        "-o",
        type=Path,
        default=".",
        help="directory to output pprox files (default current directory)",
    )
    p.add_argument(
        "--mua",
        action="store_true",
        help="save multiunit clusters along with single units",
    )
    p.add_argument(
        "--name",
        "-n",
        help="base name of the unit (default is based on 'recording' field of trials pprox) ",
    )
    p.add_argument(
        "--save-waveforms",
        "-W",
        action="store_true",
        help="save representative waveforms from each unit's main channel in npy format",
    )
    p.add_argument(
        "--waveform-num-spikes",
        type=int,
        default=200,
        help="maximum number of spikes to use for average waveform",
    )
    p.add_argument(
        "--waveform-seed",
        type=int,
        default=12345,
        help="random seed for selecting spikes",
    )
    p.add_argument(
        "--waveform-pre-peak",
        type=float,
        default=2.0,
        help="samples before the spike to keep (default %(default).1f ms)",
    )
    p.add_argument(
        "--waveform-post-peak",
        type=float,
        default=5.0,
        help="samples after the spike to keep (default %(default).1f ms)",
    )
    p.add_argument(
        "--waveform-upsample",
        type=int,
        default=3,
        help="factor to upsample spikes before aligning them (default %(default)0.1f)",
    )
    p.add_argument("recording", help="ARF recording (local file, or neurobank id/URL)")
    p.add_argument(
        "sortdir",
        type=Path,
        help="kilosort output directory. Needs to contain 'spike_times.npy', 'spike_clusters.npy',"
        " 'cluster_info.tsv', and 'temp_wh.dat'",
    )
    args = p.parse_args(argv)
    setup_log(log, args.debug)

    datafile, resource_info = get_or_verify_datafile(args.recording, args.debug)
    recording_name = resource_info["name"]
    resource_url = nbank.full_url(recording_name)
    log.info("- kilosort output directory: %s", args.sortdir)
    timefile = args.sortdir / "spike_times.npy"
    clustfile = args.sortdir / "spike_clusters.npy"
    infofile = args.sortdir / "cluster_info.tsv"
    log.info("  - spike times: %s", timefile)
    log.info("  - spike clusters: %s", clustfile)
    events = pd.DataFrame(
        {"time": np.load(timefile).squeeze(), "clust": np.load(clustfile)},
    ).set_index("clust")
    log.info("  - cluster info: %s", infofile)
    info = pd.read_csv(infofile, sep="\t", index_col=0)
    recfile = args.sortdir / "temp_wh.dat"
    params = read_kilo_params(args.sortdir / "params.py")
    recording = np.memmap(recfile, mode="c", dtype=params["dtype"])
    recording = np.reshape(
        recording, (recording.size // params["nchannels"], params["nchannels"])
    )
    nsamples, nchannels = recording.shape
    log.info("  - filtered recording: %s", recfile)
    log.info("    - %d samples, %d channels", nsamples, nchannels)
    if args.cluster is not None:
        log.info("- only analyzing clusters: %s", args.cluster)
        events = events.loc[args.cluster]

    if args.toelis:
        import toelis
        clusters = assign_events_flat(events, params["sampling_rate"])
        outfile = (args.output / recording_name).with_suffix(".toe_lis")
        if not args.dry_run:
            with open(outfile, "wt") as ofp:
                toelis.write(ofp, clusters)
            log.info("- saved %d spikes to '%s'", toelis.count(clusters), outfile)
        return

    log.info("- splitting recording into trials:")
    with h5.File(datafile, "r") as afp:
        trials = pd.DataFrame(
            oeaudio_to_trials(afp, args.sync, args.sync_thresh, args.prepad)
        )
        entry_attrs = tuple(
            entry_metadata(e) for _, e in iter_entries(afp)
        )

    # this pandas magic sorts the events by cluster and trial
    log.info("- sorting events into trials:")
    events["trial"] = trials.recording_start.searchsorted(events.time, side="left") - 1
    clusters = (
        events
        .groupby("clust")
        .apply(lambda df: df.groupby("trial").apply(lambda x: x.time.to_numpy()))
        .unstack("clust")
        .unstack("clust")         # I have no idea why this is necessary
    )

    total_spikes = 0
    total_clusters = 0
    for clust_id, cluster in clusters.items():
        clust_info = info.loc[clust_id]
        clust_type = clust_info["group"]
        clust_trials = trials.join(cluster.rename("events"))
        n_spikes = clust_trials.events.apply(len).agg("sum")
        if clust_type == "noise" or (clust_type == "mua" and not args.mua):
            log.info(
                "  - cluster %d (%d spikes, %s) -> skipped",
                clust_id,
                n_spikes,
                clust_type,
            )
            continue
        total_spikes += n_spikes
        total_clusters += 1
        outfile = args.output / f"{recording_name}_c{clust_id}.pprox"
        log.info(
            "  ✓ cluster %d (%d spikes, %s) -> %s",
            clust_id,
            n_spikes,
            clust_type,
            outfile,
        )

        clust_trials = pprox.from_trials(
            trials_to_pprox(clust_trials, params["sampling_rate"]),
            recording=resource_url,
            processed_by=["{} {}".format(p.prog, __version__)],
            kilosort_amplitude=clust_info["Amplitude"],
            kilosort_contam_pct=clust_info["ContamPct"],
            kilosort_source_channel=clust_info["ch"],
            kilosort_probe_depth=clust_info["depth"],
            kilosort_n_spikes=clust_info["n_spikes"],
            entry_metadata=entry_attrs,
            **resource_info["metadata"]
        )
        log.info("    - computing average spike waveform from channel %d", clust_info["ch"])
        n_before = int(args.waveform_pre_peak * params["sampling_rate"] / 1000)
        n_after = int(args.waveform_post_peak * params["sampling_rate"] / 1000)
        spikes = events.loc[clust_id]
        spikes = spikes[(spikes.time > n_before) & (spikes.time < (nsamples - n_after))]
        selected = spikes.sample(
            min(args.waveform_num_spikes, n_spikes), random_state=args.waveform_seed
        )
        times = sorted(selected.time)
        waveforms = qs.peaks(
            recording[:, clust_info["ch"]], times, n_before, n_after
        )
        if args.waveform_upsample > 1:
            # quick and dirty check for inverted signal
            mean_spike = waveforms.mean(0)
            flip = np.sign(mean_spike.min() + mean_spike.max())
            _, waveforms = qs.realign_spikes(
                times,
                waveforms * flip,
                args.waveform_upsample,
                expected_peak=n_before * args.waveform_upsample,
            )
            waveforms *= flip
        clust_trials["waveform"] = {
            "mean": waveforms.mean(0),
            "sampling_rate": params["sampling_rate"] * args.waveform_upsample,
        }
        if args.save_waveforms:
            np.save(args.output / (outfile.stem + "_spikes.npy"), waveforms)

        if not args.dry_run:
            with open(outfile, "wt") as ofp:
                json.dump(clust_trials, ofp, default=json_serializable)
    log.info(
        "- a total of %d spikes were assigned to %d clusters",
        total_spikes,
        total_clusters,
    )



# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Front-end scripts for song detector """
from pathlib import Path
import pickle
import argparse
import logging
import numpy as np

script_name = "quicksong"
__version__ = "2022.09.21"

_example_config = dict(
    spectrogram={
        "window_ms": 20.0,
        "shift_ms": 10.0,
        "frequency_range": [0.7, 20.0],
    },
    features=["total_power", "wiener_entropy"],
    target_label="x",
    smoothing={"kernel_widths": [11, 21, 51]},
    classifier={"gamma": 2, "C": 1},
    testing={"test_size": 0.4, "random_state": 45},
    intervals={"max_gap_ms": 100, "min_interval_ms": 500},
    datasets=[],
)

log = logging.getLogger("")


def make_config(args):
    import yaml

    cfg = _example_config.copy()
    for lblfile in args.label_files:
        cfg["datasets"].append(lblfile.stem)
    with open(args.config, "wt") as fp:
        yaml.dump(cfg, fp)
        log.info("created new configuration file '%s'", args.config)
    if args.train:
        args.dataset_dir = lblfile.parent
        args.model = None
        train_classifier(args)


def make_extractor(cfg, sampling_rate_khz):
    from quicksong import features
    from quicksong.streaming import STFT, OverlapSaveConvolver
    from quicksong.core import FeatureExtractor, make_hanning_kernels

    speccr = STFT(
        sampling_rate_khz,
        cfg["spectrogram"]["window_ms"],
        cfg["spectrogram"]["shift_ms"],
        cfg["spectrogram"]["frequency_range"],
    )
    feats = tuple(getattr(features, f) for f in cfg["features"])
    conv = OverlapSaveConvolver(make_hanning_kernels(cfg["smoothing"]["kernel_widths"]))
    return FeatureExtractor(speccr, feats, conv)


def train_classifier(args):
    import yaml
    import ewave
    from quicksong import core, model
    from arfx import lblio

    with open(args.config) as fp:
        cfg = yaml.safe_load(fp)
    log.debug(cfg)
    label_ids = {cfg["target_label"]: 0}
    log.info("Training song classifier:")
    log.info("- version: %s", __version__)

    X_all = []
    y_all = []
    for dset in cfg["datasets"]:
        log.info("- dataset: %s", dset)
        path = args.dataset_dir / dset
        with ewave.open(path.with_suffix(".wav"), "r") as fp:
            Fs = fp.sampling_rate / 1000.0
            signal = ewave.rescale(fp.read(), "f")
            log.info(" - recording: %d samples @ %.2f kHz", signal.size, Fs)
        extractor = make_extractor(cfg, Fs)
        feats = extractor.process_all(signal)
        log.info("   - calculated features %s", feats.shape)

        with open(path.with_suffix(".lbl"), "rt") as fp:
            out = []
            for lbl in lblio.read(fp):
                n = lbl["name"]
                try:
                    id = label_ids[n]
                except KeyError:
                    id = len(label_ids)
                    label_ids[n] = id
                out.append((id, lbl["start"] * 1000, lbl["stop"] * 1000))
            labels = np.array(
                out, dtype=[("id", "i4"), ("start", "f8"), ("stop", "f8")]
            )
            log.info(" - labels: %d segments", labels.size)
            X, y = core.make_training_dataset(
                feats, labels, cfg["spectrogram"]["shift_ms"]
            )
            log.info("  - labeled frames: %d", y.size)
            X_all.append(X)
            y_all.append(y)
    X = np.row_stack(X_all)
    y = np.concatenate(y_all)
    assert X.shape[0] == y.size
    log.info("- training classifier...")
    classifier, score = model.train_classifier(X, y, cfg["classifier"], cfg["testing"])
    log.info("  - model score: %.3f", score)
    if args.model is None:
        args.model = args.config.with_suffix(".pkl")
    with open(args.model, "wb") as fp:
        pickle.dump(
            {"config": cfg, "classifier": classifier, "version": __version__}, fp
        )
        log.info("- wrote model to '%s'", args.model)


class ArfWriter:
    """Utility class that writes to sequentially numbered entries"""

    def __init__(self, arfp, template="song_{index:04}"):
        self.arfp = arfp
        self.entry_count = len(arfp.keys())
        self.template = template

    @property
    def filename(self):
        return self.arfp.filename

    def create_entry(self, timestamp, **attributes):
        import arf

        next_entry_name = self.template.format(index=self.entry_count)
        self.entry_count += 1
        return arf.create_entry(self.arfp, next_entry_name, timestamp, **attributes)


def extract_songs(args):
    import os
    import arf
    import h5py as h5
    from quicksong.core import IntervalFinder

    os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"
    log.info("Extracting songs:")
    log.info("- version: %s", __version__)
    with open(args.model, "rb") as fp:
        log.info("- loading classifier from '%s'", args.model)
        model = pickle.load(fp)
    # TODO check model version
    classifier = model["classifier"]
    dt = model["config"]["spectrogram"]["shift_ms"]
    finder = IntervalFinder(max_gap=args.max_gap, min_duration=args.min_duration, dt=dt)
    # TODO: more flexibility with output format
    with arf.open_file(args.output, mode="a") as ofp:
        writer = ArfWriter(ofp)
        log.info("- saving songs to '%s'", args.output)
        for fname in args.input:
            with h5.File(fname, "r") as ifp:
                log.info("- processing recordings in '%s'", ifp.filename)
                last_sample = None
                buffer = []
                intervals = []
                for entry in ifp.values():
                    if not isinstance(entry, h5.Group):
                        log.info("  ✗ %s: not an entry, skipping", entry.name)
                        continue
                    try:
                        dset = entry[args.dataset_name]
                    except KeyError:
                        log.info(
                            "  ✗ %s: does not have dataset `%s`, skipping",
                            entry.name,
                            args.dataset_name,
                        )
                        continue
                    sampling_rate = dset.attrs["sampling_rate"] / 1000.0
                    entry_start = entry.attrs["jack_frame"]
                    log.debug(
                        " - %s: start=%s, end=%s, sampling_rate=%.2f kHz",
                        entry.name,
                        entry_start,
                        entry_start + dset.size,
                        sampling_rate,
                    )
                    # is this a continuation of the previous entry?
                    if last_sample is not None and last_sample > entry_start:
                        log.debug("   - continues previous entry")
                    else:
                        # process the intervals
                        extract_intervals(
                            writer,
                            buffer,
                            intervals,
                            args.pad_before,
                            args.pad_after,
                        )
                        extractor = make_extractor(model["config"], sampling_rate)
                        finder.reset()
                        buffer = []
                        intervals = []
                    for block in extractor.process(dset):
                        pred = classifier.predict(block)
                        intervals.extend(
                            (start * dt, stop * dt)
                            for start, stop in finder.process(pred)
                        )
                    # jack_frame is a np.uint32 so it should overflow to allow
                    # comparison with next entry
                    last_sample = entry_start + dset.size
                    buffer.append(dset)


def extract_intervals(writer, dsets, intervals, pad_before, pad_after):
    import posixpath
    from datetime import timedelta
    from arf import timestamp_to_datetime, create_dataset, DataTypes, set_uuid
    from quicksong.core import pad_intervals
    from dlab.util import all_same

    fields_to_drop = (
        "entry_creator",
        "timestamp",
        "jack_frame",
        "jack_sampling_rate",
        "jack_usec",
        "trial_off",
    )

    if len(dsets) == 0:
        return
    dset_name = ", ".join(dset.parent.name for dset in dsets)
    if len(intervals) == 0:
        log.info("  ✗ %s: no song detected", dset_name)
        return
    sampling_rate = all_same(dset.attrs["sampling_rate"] for dset in dsets) / 1000.0
    assert sampling_rate is not None, "Entries don't have the same sampling rate!"
    parent_entry = dsets[0].parent
    entry_attrs = {
        k: v for k, v in parent_entry.attrs.items() if k not in fields_to_drop
    }
    entry_attrs.update(
        source_file=parent_entry.file.filename,
        source_uuid=b",".join(dset.parent.attrs["uuid"] for dset in dsets),
        entry_creator=f"org.meliza.dlab/{script_name} {__version__}",
    )
    entry_attrs.pop("uuid")
    dset_attrs = dict(dsets[0].attrs)
    dset_attrs["source_uuid"] = b",".join(dset.attrs["uuid"] for dset in dsets)
    dset_attrs["datatype"] = DataTypes.ACOUSTIC
    dset_attrs.pop("uuid")
    signal = np.concatenate(dsets)
    log.info("  ✓ %s: ", dset_name)
    for start, end in pad_intervals(intervals, pad_before, pad_after):
        start_sample = max(0, int(start * sampling_rate))
        end_sample = min(signal.size, int(end * sampling_rate))
        song_timestamp = timestamp_to_datetime(
            parent_entry.attrs["timestamp"]
        ) + timedelta(seconds=start_sample / sampling_rate / 1000)
        new_entry = writer.create_entry(song_timestamp, **entry_attrs)
        new_dset = create_dataset(
            new_entry,
            posixpath.basename(dsets[0].name),
            signal[start_sample:end_sample],
            compression="gzip",
            **dset_attrs,
        )
        set_uuid(new_dset)
        log.info(
            "    - samples %d–%d (%.1f ms) -> %s%s",
            start_sample,
            end_sample,
            (end_sample - start_sample) / sampling_rate,
            writer.filename,
            new_dset.name,
        )


def script(argv=None):
    from dlab.util import setup_log

    p = argparse.ArgumentParser(
        description="train the song detector on labeled recordings"
    )
    p.add_argument(
        "-v", "--version", action="version", version="%(prog)s " + __version__
    )
    p.add_argument("--debug", help="show verbose log messages", action="store_true")
    sub = p.add_subparsers(title="subcommands")

    pp = sub.add_parser("init", help="generate a starting configuration file")
    pp.set_defaults(func=make_config)
    pp.add_argument(
        "--train",
        action="store_true",
        help="run training with the intialized config file",
    )
    pp.add_argument(
        "config", type=Path, help="path where the configuration file should be saved"
    )
    pp.add_argument(
        "label_files",
        type=Path,
        nargs="*",
        help="label files to add to config file as training dataset",
    )

    pp = sub.add_parser("train", help="train the classifier")
    pp.set_defaults(func=train_classifier)
    pp.add_argument(
        "--dataset-dir",
        "-d",
        type=Path,
        default=Path("."),
        help="directory where training data files are stored",
    )
    pp.add_argument("config", type=Path, help="path of the configuration file")
    pp.add_argument(
        "model",
        type=Path,
        nargs="?",
        help="path where the model should be saved (default is `config.pkl`)",
    )

    pp = sub.add_parser(
        "extract", help="use a trained model to extract songs from ARF files"
    )
    pp.set_defaults(func=extract_songs)
    pp.add_argument(
        "--dataset-name",
        default="pcm_000",
        help="the name of the dataset to process in each entry (default `%(default)s`)",
    )
    pp.add_argument(
        "--max-gap",
        type=float,
        default=100,
        help="merge segments separated by gaps less than this value (default %(default).1f ms)",
    )
    pp.add_argument(
        "--min-duration",
        type=float,
        default=500,
        help="segments must be at least this long (default %(default).1f ms)",
    )
    pp.add_argument(
        "--pad-before",
        type=float,
        default=1000,
        help="the amount of time before song onset to extract (default %(default).1f ms)",
    )
    pp.add_argument(
        "--pad-after",
        type=float,
        default=1000,
        help="the amount of time after song end to extract (default %(default).1f ms)",
    )
    pp.add_argument("model", help="trained classifier model (pkl file saved by `train`")
    pp.add_argument("input", help="the ARF file to process", nargs="+")
    pp.add_argument("output", help="output where the song segments should be saved")

    args = p.parse_args(argv)
    setup_log(log, args.debug)
    if not hasattr(args, "func"):
        p.print_usage()
        return 0

    args.func(args)


if __name__ == "__main__":
    script()



# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Commandline script to extract waveforms from extracellular recordings """
import json
import logging
import nbank
import numpy as np
import h5py as h5
import pandas as pd
import quickspikes as qs

log = logging.getLogger("dlab")


#def script(argv=None):
if __name__=="__main__":
    import argparse
    from dlab.core import get_or_verify_datafile
    from dlab.util import setup_log

    __version__ = "2022.06.24"

    p = argparse.ArgumentParser(
        description="extract waveforms from extracellular recording based "
        "on spike times in pprox files. A csv file is generated"
        "for each pprox file"
    )
    p.add_argument(
        "-v", "--version", action="version", version="%(prog)s " + __version__
    )
    p.add_argument("--debug", help="show verbose log messages", action="store_true")
    p.add_argument(
        "--num-spikes",
        "-n",
        type=int,
        default=100,
        help="maximum number of spikes to extract",
    )
    p.add_argument(
        "--pre-spike",
        type=float,
        default=2.0,
        help="samples before the spike to keep (default %(default).1f ms)",
    )
    p.add_argument(
        "--post-spike",
        type=float,
        default=5.0,
        help="samples after the spike to keep (default %(default).1f ms)",
    )
    p.add_argument(
        "--upsample",
        type=float,
        default=3,
        help="factor to upsample spikes before aligning them (default %(default)0.1f)",
    )
    p.add_argument(
        "--seed",
        type=int,
        default=12345,
        help="random seed for selecting spikes"
    )
    p.add_argument("recording", help="ARF recording (local file, or neurobank id/URL)")
    p.add_argument(
        "pprox",
        nargs="+",
        type=argparse.FileType("r", encoding="utf-8"),
        help="pprox files with spike times",
    )
    args = p.parse_args()
    setup_log(log, args.debug)

    datafile, arf_info = get_or_verify_datafile(args.recording, args.debug)

    with h5.File(datafile, "r") as arf:
        for spikefile in args.pprox:
            log.info(" - processing spike times in '%s':", spikefile.name)
            spike_data = json.load(spikefile)
            # check that the recording matches the ARF file
            log.info("   - source recording: %s", spike_data["recording"])
            dset_channel = spike_data["kilosort_source_channel"]
            # TODO verify that that index is 1-based
            dset_name = f"CH{dset_channel + 1}"
            log.info("   - main channel: %s", dset_name)
            resource_info = nbank.describe(spike_data["recording"])
            if resource_info["sha1"] != arf_info["sha1"]:
                log.warning(
                    "   - warning: recording does not match the supplied ARF file, skipping"
                )
                continue
            # convert spike times to sample counts (relative to entry start)
            entry_metadata = spike_data["entry_metadata"]
            if len(set(e["sampling_rate"] for e in entry_metadata)) > 1:
                raise ValueError("Not all entries have the same sampling rate")
            sampling_rate = entry_metadata[0]["sampling_rate"]
            dsets = [arf[m["name"]][dset_name] for m in entry_metadata]
            n_before = int(args.pre_spike * sampling_rate / 1000)
            n_after = int(args.post_spike * sampling_rate / 1000)
            spikes = []
            for trial in spike_data["pprox"]:
                entry = trial["recording"]["entry"]
                sampling_rate = entry_metadata[entry]["sampling_rate"]
                for spike in trial["events"]:
                    peak_sample = (
                        int(spike * sampling_rate) + trial["recording"]["start"]
                    )
                    spike_start = peak_sample - n_before
                    spike_end = peak_sample + n_after
                    if spike_start > 0 and spike_end < trial["recording"]["stop"]:
                        spikes.append(
                            (entry, spike, spike_start, spike_end)
                        )
            spikes = pd.DataFrame.from_records(
                np.array(spikes, dtype=[("entry", "i4"), ("peak", "f8"), ("start", "i8"), ("stop", "i8")])
            )
            nspikes, _ = spikes.shape
            selected = spikes.sample(min(args.num_spikes, nspikes), random_state=args.seed)
            nspikes, _ = selected.shape

            waveforms = np.zeros((nspikes, n_before + n_after), dtype="f8")
            for i, row in enumerate(selected.itertuples(index=False)):
                waveforms[i, :] = dsets[row.entry][row.start:row.stop]
            #__import__("IPython").embed()


            #     spike_idx = (np.asarray(trial["events"]) * sampling_rate).astype("int64") + trial["recording"]["start"]
            #     nspikes = spike_idx.size
            #     log.debug("    - trial %d: %d spikes from entry %d", trial["index"], nspikes, entry)
            #     spike_times.append(spike_idx)
            #     spike_entries.append(np.ones(nspikes, dtype="int32") * entry)
            # all_spike_entries = np.concatenate(spike_entries)
            # all_spike_times = np.concatenate(spike_times)
            # nspikes = all_spike_times.size

            # TODO sort by entry to avoid looking up the dataset on every spike
            # spike_tbl = pd.DataFrame({"entry":
            # spikes = []
            # for entry, peak in zip(all_spike_entries[selected], all_spike_times[selected]):
            #     
            #     s_before = peak - args.pre_spike * entry_metadata[entry]["sampling_rate"]
            #     s_after = peak + args.post_spike * entry_metadata[entry]["sampling_rate"]
            #     spikes.append(dset[s_before:s_after])



# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Utilities for extracellular experiments """
import os
import re
import json
import logging

import quickspikes as qs
import nbank
import numpy as np
import h5py as h5
from dlab import pprox
from dlab.util import memodict

log = logging.getLogger("dlab")


### general


def entry_time(entry):
    from arf import timestamp_to_float

    return timestamp_to_float(entry.attrs["timestamp"])


def iter_entries(data_file):
    """Iterate through the entries in an arf file in order of time"""
    return enumerate(sorted(data_file.values(), key=entry_time))


### present-audio


def audiolog_to_trials(trials, data_file, sync_dset="channel37", sync_thresh=1.0):
    """Parses a logfile from Margot's present_audio scripts for experiment structure

    trials: the "presentation" field in the json output of present_audio.py
    data_file: open handle to the hdf5 file generated by open-ephys during the recording
    sync_dset: the name of the dataset containing the synchronization signal
    sync_thresh: the threshold for detecting the sync signal
    trials: number of trials per stimulus
    """
    from arf import timestamp_to_float

    # Each element in this structure corresponds to a trial. In some cases the
    # data are stored as a dict/map, but the keys are just strings of the trial
    # number. The indices correspond to the entries in the arf file.
    n_trials = len(trials)
    expt_start = None
    sample_count = 0
    det = qs.detector(sync_thresh, 10)
    for i in range(n_trials):
        pproc = {"events": [], "index": i}
        pproc.update(trials[str(i)])
        entry_name = "/rec_%d" % i
        entry = data_file[entry_name]

        # get time relative to first trial
        time = entry_time(entry)
        if expt_start is None:
            expt_start = time
        pproc["offset"] = time - expt_start
        # find the sync signal - we expect one and only one click
        dset = entry[sync_dset]
        pproc["recording"] = {
            "entry": entry_name,
            "start": int(sample_count),
            "stop": int(sample_count + dset.size),
            "sampling_rate": dset.attrs["sampling_rate"],
        }
        sample_count += dset.size
        sync = dset[:].astype("d")
        det.scale_thresh(sync.mean(), sync.std())
        clicks = det(sync)
        if len(clicks) != 1:
            log.error("%s: expected 1 click, detected %d", dset.name, len(clicks))
        else:
            pproc["stim_on"] = clicks[0] / dset.attrs["sampling_rate"]
        yield pproc


def audiolog_to_pprox_script(argv=None):
    """CLI to generate a pprox from present_audio log"""
    import sys
    import argparse
    import json
    from dlab.util import setup_log, json_serializable

    __version__ = "0.1.0"

    p = argparse.ArgumentParser(
        description="generate pprox from trial structure in present_audio logfile"
    )
    p.add_argument(
        "-v", "--version", action="version", version="%(prog)s " + __version__
    )
    p.add_argument("--debug", help="show verbose log messages", action="store_true")
    p.add_argument(
        "--output",
        "-o",
        type=argparse.FileType("w", encoding="utf-8"),
        default=sys.stdout,
        help="name of output file. If absent, outputs to stdout",
    )
    p.add_argument(
        "--sync",
        default="channel37",
        help="name of channel with synchronization signal",
    )
    p.add_argument(
        "--sync-thresh",
        default="30.0",
        type=float,
        help="threshold (z-score) for detecting synchronization clicks",
    )
    p.add_argument("logfile", help="log file generated by present_audio.py")
    p.add_argument("recording", help="neurobank id or URL for the ARF recording")
    args = p.parse_args(argv)
    setup_log(log, args.debug)

    resource_url = nbank.full_url(args.recording)
    datafile = nbank.get(args.recording, local_only=True)
    if datafile is None:
        p.error(
            "unable to locate resource %s - is it deposited in neurobank?"
            % args.recording
        )
    log.info("loading recording from resource %s", resource_url)

    with h5.File(datafile, "r") as afp:
        with open(args.logfile, "rt") as lfp:
            expt_log = json.load(lfp)
            trials = pprox.from_trials(
                audiolog_to_trials(
                    expt_log.pop("presentation"), afp, args.sync, args.sync_thresh
                ),
                recording=resource_url,
                processed_by=["{} {}".format(p.prog, __version__)],
                **expt_log
            )
    json.dump(trials, args.output, default=json_serializable)
    if args.output != sys.stdout:
        log.info("wrote trial data to '%s'", args.output.name)


############### open-ephys-audio specific functions

def find_stim_dset(entry):
    """Returns the first dataset that matches 'Network_Events.*_TEXT'"""
    rex = re.compile(r"Network_Events-.*?TEXT")
    for name in entry:
        if rex.match(name) is not None:
            log.debug("  - stim log dataset: %s", name)
            return entry[name]


def parse_stim_id(path):
    """Extracts the stimulus id from the path"""
    return os.path.splitext(os.path.basename(path))[0]


@memodict
def stim_duration(stim_name):
    """
    Returns the duration of a stimulus (in s). This can only really be done by
    downloading the stimulus from the registry, because the start/stop times are
    not reliable. We try to speed this up by memoizing the function and caching
    the downloaded files.

    """
    import wave
    from dlab.core import fetch_resource

    neurobank_registry = nbank.default_registry()
    target = fetch_resource(neurobank_registry, stim_name)
    with open(target, "rb") as fp:
        reader = wave.open(fp)
        return 1.0 * reader.getnframes() / reader.getframerate()


def entry_metadata(entry):
    """Extracts metadata from an entry in an oeaudio-present experiment ARF file.

    Metadata are passed to open-ephys through the network events socket as a
    json-encoded dictionary. There should be one and only one metadata message
    per entry, so only the first is returned.

    """
    re_metadata = re.compile(r"metadata: (\{.*\})")
    stims = find_stim_dset(entry)
    for row in stims:
        message = row["message"].decode("utf-8")
        m = re_metadata.match(message)
        try:
            metadata = json.loads(m.group(1))
        except (AttributeError, json.JSONDecodeError):
            pass
        else:
            metadata.update(name=entry.name, sampling_rate=stims.attrs["sampling_rate"])
            return metadata



# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Functions for processing acoustic data """

import os
import numpy as np
import h5py as h5
import libtfr


os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"

__version__ = "20220808-1"




# -*- coding: utf-8 -*-
# -*- mode: python -*-
"""Process pprox (point process) data

See https://meliza.org/spec:2/pprox/ for specification

This package deliberately does not include any functions for serialization, as
pprox objects are just python dictionaries.

Copyright (C) Dan Meliza, 2006-2020 (dan@meliza.org)

"""

_schema = "https://meliza.org/spec:2/pprox.json#"


def empty():
    """Returns a new, empty pprox object"""
    return from_trials([])


def from_trials(trials, **metadata):
    """Wrap a sequence of trials in a pprox object, optionally specifying top-level metadata"""
    d = { "$schema": _schema, "pprox": tuple(trials) }
    d.update(**metadata)
    return d


def wrap_uuid(b):
    """ Wrap a UUID (string or bytes) in a URN string """
    import uuid
    try:
        b = b.decode("ascii")
    except AttributeError:
        pass
    return uuid.UUID(b).urn


def groupby(obj, *keys):
    """Iterate through pprocs based on keys

    For example, if "stim" and "trial" are metadata on the trials, collate(obj, "stim") will yield
    (stim0, [events_trial0, events_trial1]), (stim1, [events_trial0, events_trial1]), ...

    """
    import itertools
    import operator
    keyfun = operator.itemgetter(keys)
    evsorted = sorted(obj['pprox'], keyfun)
    return itertools.groupby(evsorted, keyfun)


def validate(obj):
    """Validates object against pprox schema"""
    import jsonschema
    import requests
    pass


def trial_iterator(pprox):
    """ Iterates through trials in pprox, yielding (index, trial) """
    for i, trial in enumerate(pprox["pprox"]):
        # annotate with sampling rate
        if "sampling_rate" not in trial["recording"]:
            entry = trial["recording"]["entry"]
            sampling_rate = pprox["entry_metadata"][entry]["sampling_rate"]
            trial["recording"]["sampling_rate"] = sampling_rate
        yield i, trial


def aggregate_events(pprox, use_recording=False):
    """Aggregate all the events in a pprox into a single array.

    This function is primarily used for testing, as this should be the reverse
    operation to whatever function is used to assign events to trials (assuming
    no gaps in the recording). If there are gaps, then set `use_recording` to
    True.

    """
    import numpy as np
    all_events = []
    for trial in pprox["pprox"]:
        sampling_rate = trial["recording"]["sampling_rate"]
        events = np.asarray(trial["events"])
        if use_recording:
            events = (events * sampling_rate).astype("i8") + trial["recording"]["start"]
        else:
            events = ((events + trial["offset"]) * sampling_rate).astype("i8")
        all_events.append(events)
    return np.concatenate(all_events)
